
#include <wx/wx.h>
#include <wx/file.h>
#include <memory.h>
#include "lodeframe.h"
#include "app.h"
#include "offsetdlg.h"



IMPLEMENT_APP(LodeApp);


LodeApp::LodeApp() : wxApp()
{
	bFileLoaded = 0;
	bIsDirty = 0;
}

bool LodeApp::OnInit()
{
	wxApp::OnInit();

	wFrame = new LodeFrame(NULL,wxID_ANY,wxT("Lode Funner"));
	wFrame->Show();
	SetTopWindow(wFrame);

	return true;
}

int LodeApp::SavePrompt()
{
	if(!bFileLoaded)
		return 0;

	if(!bIsDirty)
		return 0;

	int what = wxMessageBox(wxT("Save changes before proceeding?"),wxT("Save Changes?"),
		wxYES_NO | wxCANCEL | wxICON_QUESTION,wFrame);

	switch(what)
	{
	case wxYES:		return Save();
	case wxNO:		return 0;
	case wxCANCEL:	return 1;
	}

	return 1;
}

int LodeApp::Save()
{
	if(!bFileLoaded)
		return 0;

	int needed = DoCalc();
	if(needed > (nMaxBytes*8))
	{
		wxMessageBox(wxString::Format(wxT("Maps cannot be compressed to a small enough size.\nMaps compress to %d bits.\nMaximum is %d bits."),needed,nMaxBytes * 8),
			wxT("Cannot Save"),wxOK | wxICON_ERROR,wFrame);
		return 1;
	}

	wxFile file;
	if(!file.Open(strFilePath,wxFile::read_write))
	{
		wxMessageBox(wxT("Unable to open ROM for writing.\nMake sure ROM is not marked as read-only."),
			wxT("Cannot Save"),wxOK | wxICON_ERROR,wFrame);
		return 1;
	}

	file.Seek(nStartOffset);

	u8* buf = new u8[nMaxBytes];
	memset(buf,0,nMaxBytes);

	needed = (levels.Save(buf,nMaxBytes) + 7) / 8;

	file.Write(buf,needed);
	file.Close();

	delete[] buf;
	bIsDirty = 0;

	return 0;
}

int LodeApp::DoCalc()
{
	return levels.Save(0,0);
}

void LodeApp::Calc()
{
	if(!bFileLoaded)
		return;

	int bits = DoCalc();
	wxMessageBox(wxString::Format(wxT("Maps compress to %d bits.\nMaximum is %d bits."),bits,nMaxBytes * 8),
		wxT("Bit Calculation"),wxOK | wxICON_INFORMATION,wFrame);
}

int LodeApp::Load()
{
	wxString fn = wxFileSelector(wxT("Select a ROM to open"),wxEmptyString,wxEmptyString,wxEmptyString,
		wxT("NES ROMs (*.nes)|*.nes|All Files|*"),wxFD_FILE_MUST_EXIST,wFrame);

	if(fn.IsEmpty())
		return 1;

	if(SavePrompt())
		return 1;

	// load the file
	wxFile file;
	if(!file.Open(fn,wxFile::read))
		return 1;

	OffsetDlg dlg(wFrame,wxID_ANY,wxT("File Offsets"));

	GetConfigOffsets(fn,dlg.nOffsets);

	if(dlg.ShowModal() != wxID_OK)
		return 1;
	
	SetConfigOffsets(fn,dlg.nOffsets);

	nMaxBytes = dlg.nOffsets[1] - dlg.nOffsets[0];
	nStartOffset = dlg.nOffsets[0];

	unsigned char palbuf[0x20];
	unsigned char chrbuf[0x280];
	unsigned char* levbuf = new unsigned char[nMaxBytes];

	memset(palbuf,0,0x20);
	memset(chrbuf,0,0x280);
	memset(levbuf,0,nMaxBytes);

	file.Seek(dlg.nOffsets[3]);
	file.Read(palbuf,0x20);

	file.Seek(dlg.nOffsets[2]);
	file.Read(chrbuf,0x280);
	
	file.Seek(dlg.nOffsets[0]);
	file.Read(levbuf,nMaxBytes);

	file.Close();

	// get graphics
	gfx.Load(chrbuf,palbuf);

	// get maps
	levels.Load(levbuf,nMaxBytes);

	delete[] levbuf;

	// this file is loaded
	bFileLoaded = 1;
	bIsDirty = 0;
	strFilePath = fn;
	wFrame->FileLoaded(fn);

	return 0;
}

void LodeApp::SetTile(int tile)
{
	wFrame->SetTile(tile);
}

void LodeApp::ImportLevel()
{
	if(!bFileLoaded)
		return;
	
	wxString fn = wxFileSelector(wxT("Select a Level to import"),wxEmptyString,wxEmptyString,wxEmptyString,
		wxT("Exported Levels (*.lode)|*.lode|All Files|*"),wxFD_FILE_MUST_EXIST,wFrame);

	if(fn.IsEmpty())
		return;

	// load the file
	wxFile file;
	if(!file.Open(fn,wxFile::read))
	{
		::wxMessageBox(wxT("Unable to load exported level."),wxT("Error"),wxOK | wxICON_ERROR,wFrame);
		return;
	}

	unsigned char buf[(28*14) + (8*2)] = {0};
	file.Read(buf,(28*14)+(8*2));
	file.Close();
	
	int i;
	for(i = 0; i < (28*14); ++i)
		buf[i] &= 7;

	LEVEL& lev = levels.Get(wFrame->GetLevel());
	memcpy(lev.mp[0],buf,(28*14));

	for(i = 0; i < 7; ++i)
	{
		lev.enX[i] = buf[(28*14)+(2*i)];
		lev.enY[i] = buf[(28*14)+(2*i)+1];

		if((lev.enX[i] >= 28) || (lev.enY[i] >= 14))
			break;
	}
	lev.enemies = (u8)i;

	lev.plX = buf[(28*14)+(7*2)];
	lev.plY = buf[(28*14)+(7*2)+1];

	if(lev.plX >= 28)	lev.plX = 27;
	if(lev.plY >= 14)	lev.plY = 13;

	if(!lev.enemies)
	{
		lev.enemies = 1;
		lev.enX[0] = 0;
		lev.enY[0] = 0;
	}

	if((lev.enemies == 1) && (lev.enX[0] == lev.plX) && (lev.enY[0] == lev.plY))
	{
		if(lev.enX[0])		--lev.enX[0];
		else				++lev.enX[0];
	}

	bIsDirty = 1;
	wFrame->Redraw();
}

void LodeApp::ExportLevel()
{
	if(!bFileLoaded)
		return;
	
	wxString fn = wxFileSelector(wxT("Export Level"),wxEmptyString,wxEmptyString,wxEmptyString,
		wxT("Exported Levels (*.lode)|*.lode|All Files|*"),0,wFrame);

	if(fn.IsEmpty())
		return;

	// open the file
	wxFile file;
	if(!file.Open(fn,wxFile::write))
	{
		::wxMessageBox(wxT("Unable to save exported level."),wxT("Error"),wxOK | wxICON_ERROR,wFrame);
		return;
	}

	unsigned char buf[(28*14) + (8*2)];

	LEVEL& lev = levels.Get(wFrame->GetLevel());
	memcpy(buf,lev.mp[0],(28*14));
	
	int i;
	for(i = 0; i < lev.enemies; ++i)
	{
		buf[(28*14)+(2*i)] = lev.enX[i];
		buf[(28*14)+(2*i)+1] = lev.enY[i];
	}
	for(; i < 7; ++i)
	{
		buf[(28*14)+(2*i)] = 0xFF;
		buf[(28*14)+(2*i)+1] = 0xFF;
	}

	buf[(28*14)+(7*2)] = lev.plX;
	buf[(28*14)+(7*2)+1] = lev.plY;

	file.Write(buf,(28*14)+(8*2));
	file.Close();
}