
#include <wx/wx.h>
#include <wx/dc.h>
#include <wx/dcmemory.h>
#include <wx/image.h>
#include <wx/bitmap.h>
#include "graphic.h"
#include <malloc.h>

static const unsigned long NES_PALETTE[0x40] =
{
0x788084,0x0000FC,0x0000C4,0x4028C4,0x94008C,0xAC0028,0xAC1000,0x8C1800,
0x503000,0x007800,0x006800,0x005800,0x004058,0x000000,0x000000,0x000000,
0xBCC0C4,0x0078FC,0x0088FC,0x6848FC,0xDC00D4,0xE40060,0xFC3800,0xE46018,
0xAC8000,0x00B800,0x00A800,0x00A848,0x008894,0x000000,0x000000,0x000000,
0xFCF8FC,0x38C0FC,0x6888FC,0x9C78FC,0xFC78FC,0xFC589C,0xFC7858,0xFCA048,
0xFCB800,0xBCF818,0x58D858,0x58F89C,0x00E8E4,0x606060,0x000000,0x000000,
0xFCF8FC,0xA4E8FC,0xBCB8FC,0xDCB8FC,0xFCB8FC,0xF4C0E0,0xF4D0B4,0xFCE0B4,
0xFCD884,0xDCF878,0xB8F878,0xB0F0D8,0x00F8FC,0xC8C0C0,0x000000,0x000000
};


Graphic::Graphic()
{
	Zero();
}

Graphic::~Graphic()
{
	Destroy();
}

void Graphic::Zero()
{
	pImg = 0;
	pBmp = 0;
	pDC = 0;
	
	pDblImg = 0;
	pDblBmp = 0;
	pDblDC = 0;
}

void Graphic::Load(const unsigned char* buf,const unsigned char* palettes)
{
	Destroy();
	
	unsigned char* pixels = (unsigned char*)malloc(3 * 16 * 16 * 10);
	// fill pixels

	static const int tilepal[10] = {1,1,1,2,2,1,2,3,5,4};

	int offset = 0;
	int i;
	unsigned long clr;
	int x, y;
	unsigned char* p;

	for(i = 0; i < (10*4); ++i, offset += 0x10)
	{
		p = pixels + ((i>>1) * 3*8);
		if(i & 1)		p += (3*16*10*8);

		for(y = 0; y < 8; ++y, p += (3*16*10))
		{
			for(x = 0; x < 8; ++x)
			{
				clr =  (buf[offset+y] >> (7-x)) & 1;
				clr |= ((buf[offset+y+8] << 1) >> (7-x)) & 2;
				if(clr)		clr = palettes[(tilepal[i/4] * 4) + clr];
				else		clr = palettes[0];

				clr = NES_PALETTE[clr & 0x3F];

				p[(x*3)+0] = (clr >> 16) & 0xFF;		//r
				p[(x*3)+1] = (clr >>  8) & 0xFF;		//g
				p[(x*3)+2] = (clr      ) & 0xFF;		//b
			}
		}
	}
	
	// do the doubled image
	p = (unsigned char*)malloc(3 * 16 * 16 * 10 * 4);
	unsigned char* a = pixels;
	unsigned char* b = p;
	unsigned char* c = p + (3*32*10);

	for(y = 0; y < 16; ++y)
	{
		for(x = 0; x < 10*16; ++x)
		{
			b[0] = b[3] = c[0] = c[3] = a[0];
			b[1] = b[4] = c[1] = c[4] = a[1];
			b[2] = b[5] = c[2] = c[5] = a[2];

			a += 3;
			b += 6;
			c += 6;
		}
		b += (3*32*10);
		c += (3*32*10);
	}
	pDblImg = new wxImage(32*10,32,false);
	pDblImg->SetData(p);

	// create our image
	pImg = new wxImage(16*10,16,false);
	pImg->SetData(pixels);

	// put image in bitmap, and that bitmap in our mem DC
	pBmp = new wxBitmap(*pImg);
	pDC = new wxMemoryDC(*pBmp);

	pDblBmp = new wxBitmap(*pDblImg);
	pDblDC = new wxMemoryDC(*pDblBmp);
}

void Graphic::Destroy()
{
	if(pDC)		delete pDC;
	if(pBmp)	delete pBmp;
	if(pImg)	delete pImg;
	if(pDblDC)	delete pDblDC;
	if(pDblBmp)	delete pDblBmp;
	if(pDblImg)	delete pDblImg;
	Zero();
}

void Graphic::DrawTile(int tile,wxDC* dc,int x,int y,int stretch)
{
	if(!pDC)		return;

	if(stretch)
		dc->Blit(x*32,y*32,32,32,pDblDC,tile*32,0);
	else
		dc->Blit(x*16,y*16,16,16,pDC,tile*16,0);
}
