
#include <memory.h>
#include "levels.h"

/*
 *	Bit Packer
 */

BitPacker::BitPacker(u8* buf,int mxlen)
{
	pBuf =		buf;
	nPos =		0;
	nMaxLen =	mxlen;
	nBit =		0;
}

u8 BitPacker::Get1()
{
	if(nPos >= nMaxLen)
	{
		IncBit();
		return 0;
	}

	u8 ret;
	ret = (pBuf[nPos] & (0x80 >> nBit)) ? 1 : 0;

	IncBit();
	return ret;
}

u8 BitPacker::Get2()
{
	u8 ret = Get1() << 1;
	ret |= Get1();
	return ret;
}

u8 BitPacker::Get3()
{
	u8 ret = Get1() << 2;
	ret |= Get1() << 1;
	ret |= Get1();
	return ret;
}

u8 BitPacker::Get5()
{
	u8 ret = Get1() << 4;
	ret |= Get1() << 3;
	ret |= Get1() << 2;
	ret |= Get1() << 1;
	ret |= Get1();
	return ret;
}

void BitPacker::IncBit()
{
	++nBit;
	if(nBit >= 8)
	{
		++nPos;
		nBit = 0;
	}
}

void BitPacker::Set1(u8 v)
{
	if(nPos >= nMaxLen)
	{
		IncBit();
		return;
	}

	if(v)		pBuf[nPos] |=  (0x80 >> nBit);
	else		pBuf[nPos] &= ~(0x80 >> nBit);

	IncBit();
}

void BitPacker::Set2(u8 v)
{
	Set1(v & 0x02);
	Set1(v & 0x01);
}

void BitPacker::Set3(u8 v)
{
	Set1(v & 0x04);
	Set1(v & 0x02);
	Set1(v & 0x01);
}

void BitPacker::Set5(u8 v)
{
	Set1(v & 0x10);
	Set1(v & 0x08);
	Set1(v & 0x04);
	Set1(v & 0x02);
	Set1(v & 0x01);
}


/*
 *	The Levels
 */

Levels::Levels()
{
}

Levels::~Levels()
{
}

void Levels::Zero()
{
	memset(lvl,0,sizeof(LEVEL) * 50);
	int i;
	for(i = 0; i < 50; ++i)
	{
		lvl[0].plY = 1;
		lvl[0].enemies = 1;
	}
}

void Levels::Load(u8* buf,int mxlen)
{
	Zero();

	BitPacker	bt(buf,mxlen);

	int lv, i;
	u8 x, y;
	u8 ctl;
	u8 tl;
	u8 pl[4];

	for(lv = 0; lv < 50; ++lv)
	{
		for(y = 0; y < 14; ++y)
		{
			x = 0;
			switch(bt.Get2())
			{
			case 0:
				/*
				 *	RLE sets
				 */
				ctl = bt.Get1();
				while(ctl)
				{
					i = bt.Get5();
					tl = bt.Get3();
					ctl = bt.Get1();

					if(!i)		i = 28;
					for(; (i > 0) && (x < 28); --i, ++x)
						lvl[lv].mp[y][x] = tl;
				}
				tl = bt.Get3();
				for(; x < 28; ++x)
					lvl[lv].mp[y][x] = tl;
				break;

			case 1:
				/*
				 *	1 bit per tile "bitmap"
				 */
				pl[0] = bt.Get3();
				pl[1] = bt.Get3();
				for(x = 0; x < 28; ++x)
					lvl[lv].mp[y][x] = pl[bt.Get1()];
				break;

			case 2:
				/*
				 *	All tiles 0 except for...
				 */
				memset(lvl[lv].mp[y],0,28);
				ctl = bt.Get1();
				while(ctl)
				{
					tl = bt.Get3();
					i = bt.Get5();
					ctl = bt.Get1();
					if(i < 28)
						lvl[lv].mp[y][i] = tl;
				}
				break;

			case 3:
				if(bt.Get1())
				{
					/*
					 *	uncompressed
					 */
					for(x = 0; x < 28; ++x)
						lvl[lv].mp[y][x] = bt.Get3();
				}
				else
				{
					
					/*
					 *	2 bit per tile "bitmap"
					 */
					pl[0] = bt.Get3();
					pl[1] = bt.Get3();
					pl[2] = bt.Get3();
					pl[3] = bt.Get3();
					for(x = 0; x < 28; ++x)
						lvl[lv].mp[y][x] = pl[bt.Get2()];
				}
				break;
			}
		}

		lvl[lv].enemies = bt.Get3();
		if(!lvl[lv].enemies)		lvl[lv].enemies = 1;

		for(i = 0; i < lvl[lv].enemies; ++i)
		{
			x = bt.Get5();
			y = bt.Get5();
			if(x >= 28)		x = 27;
			if(y >= 14)		y = 13;

			lvl[lv].enX[i] = x;
			lvl[lv].enY[i] = y;
		}

		x = bt.Get5();
		y = bt.Get5();
		if(x >= 28)		x = 27;
		if(y >= 14)		y = 13;
		lvl[lv].plX = x;
		lvl[lv].plY = y;
	}
}



u8 Levels::Get(int lev,int x,int y)
{
	if((x == lvl[lev].plX) && (y == lvl[lev].plY))
		return 9;

	int i;
	for(i = 0; i < lvl[lev].enemies; ++i)
	{
		if(x != lvl[lev].enX[i])		continue;
		if(y != lvl[lev].enY[i])		continue;

		return 8;
	}

	return lvl[lev].mp[y][x];
}

void Levels::Set(int lev,int x,int y,int tile,int& alt_x,int& alt_y)
{
	if(tile < 0)		return;
	if(tile > 9)		return;

	int en = -1;
	int i;
	for(i = 0; i < lvl[lev].enemies; ++i)
	{
		if(lvl[lev].enX[i] != x)		continue;
		if(lvl[lev].enY[i] != y)		continue;

		en = i;
		break;
	}


	if(tile == 9)		// player pos
	{
		if(KillEnemy(lev,en))
		{
			lvl[lev].enX[en] = lvl[lev].plX;
			lvl[lev].enY[en] = lvl[lev].plY;
		}

		alt_x = lvl[lev].plX;
		alt_y = lvl[lev].plY;
		lvl[lev].plX = (u8)x;
		lvl[lev].plY = (u8)y;
		return;
	}

	if(tile == 8)		// enemy
	{
		if(en >= 0)		// already an enemy here
			return;

		// can't put enemy on the player
		if((lvl[lev].plX == x) && (lvl[lev].plY == y))
			return;

		// can't have more than 7 enemies
		if(lvl[lev].enemies >= 7)
		{
			alt_x = lvl[lev].enX[6];
			alt_y = lvl[lev].enY[6];
			KillEnemy(lev,6);
		}

		lvl[lev].enX[lvl[lev].enemies] = (u8)x;
		lvl[lev].enY[lvl[lev].enemies] = (u8)y;
		++lvl[lev].enemies;
		return;
	}

	// otherwise, normal tile
	KillEnemy(lev,en);
	lvl[lev].mp[y][x] = tile;
}

int Levels::KillEnemy(int lev,int en)
{
	if(en < 0)						return 0;
	if(en >= lvl[lev].enemies)		return 0;

	if(lvl[lev].enemies == 1)		return 1;		// can't kill last enemy

	--lvl[lev].enemies;
	lvl[lev].enX[en] = lvl[lev].enX[lvl[lev].enemies];
	lvl[lev].enY[en] = lvl[lev].enY[lvl[lev].enemies];
	return 0;
}

/*
 *	Saving
 */

int Levels::Save(u8* buf,int mxlen)
{
	BitPacker		bt(buf,mxlen);

	int lev, y, x, i;
	u8 ctl;
	u8 tmp[4];
	u8 pal[8];

	for(lev = 0; lev < 50; ++lev)
	{
		for(y = 0; y < 14; ++y)
		{
			switch(FindBestRowType(lev,y,tmp))
			{
				/*
				 *	Block 0 -- RLE blocks
				 */
			case 0:
				bt.Set2(0);
				x = 0;
				for(i = 1; i < 28; ++i)
				{
					if(lvl[lev].mp[y][x] != lvl[lev].mp[y][i])
					{
						bt.Set1(1);
						bt.Set5(i-x);
						bt.Set3(lvl[lev].mp[y][x]);
						x = i;
					}
				}
				bt.Set1(0);
				bt.Set3(lvl[lev].mp[y][x]);
				break;

				/*
				 *	Block 1 -- 1bit per tile bitmap
				 */
			case 1:
				// tmp has the palette
				memset(pal,0,8);
				pal[tmp[1]] = 1;

				bt.Set2(1);
				bt.Set3(tmp[0]);
				bt.Set3(tmp[1]);
				for(x = 0; x < 28; ++x)
					bt.Set1(pal[lvl[lev].mp[y][x]]);
				break;

				/*
				 *	Block 2 -- "all zeros except..."
				 */
			case 2:
				bt.Set2(2);
				for(x = 0; x < 28; ++x)
				{
					ctl = lvl[lev].mp[y][x];
					if(ctl)
					{
						bt.Set1(1);
						bt.Set3(ctl);
						bt.Set5(x);
					}
				}
				bt.Set1(0);
				break;

				/*
				 *	Block 3A -- 2 bit per tile bitmap
				 */
			case 3:
				memset(pal,0,8);
				pal[tmp[1]] = 1;
				pal[tmp[2]] = 2;
				pal[tmp[3]] = 3;

				bt.Set2(3);
				bt.Set1(0);
				bt.Set3(tmp[0]);
				bt.Set3(tmp[1]);
				bt.Set3(tmp[2]);
				bt.Set3(tmp[3]);
				for(x = 0; x < 28; ++x)
					bt.Set2(pal[lvl[lev].mp[y][x]]);
				break;

				/*
				 *	Block 3B -- uncompressed
				 */
			case 4:
				bt.Set2(3);
				bt.Set1(1);
				for(x = 0; x < 28; ++x)
					bt.Set3(lvl[lev].mp[y][x]);
				break;
			}
		}

		bt.Set3(lvl[lev].enemies);
		for(i = 0; i < lvl[lev].enemies; ++i)
		{
			bt.Set5(lvl[lev].enX[i]);
			bt.Set5(lvl[lev].enY[i]);
		}
		
		bt.Set5(lvl[lev].plX);
		bt.Set5(lvl[lev].plY);
	}

	return bt.GetNumBits();
}

int Levels::FindBestRowType(int lev,int y,u8* pal)
{
	u8* row = lvl[lev].mp[y];

	// best type (default) is uncompressed (3B)
	int besttype = 4;
	int bestlen = 3 + (28*3);

	int len;
	int i, j, k;

	// replace enemies/player with the tile adjacent to them
	FillObjectTiles(lev,y,1);

	/*
	 *	First check to see if we can palettize this row
	 */
	pal[0] = row[0];
	j = 1;
	for(i = 1; i < 28; ++i)
	{
		for(k = 0; k < j; ++k)
		{
			if(pal[k] == row[i])	break;
		}
		if(k == j)	// tile wasn't in palette, try to put it in
		{
			++j;
			if(j > 4)	// too many different tiles -- can't palettize
				break;

			pal[k] = row[i];
		}
	}
	if(j <= 2)
	{
		besttype = 1;
		bestlen = 2+(3*2)+28;
	}
	else if(j <= 4)
	{
		besttype = 3;
		bestlen = 2+1+(3*4)+(28*2);
	}

	/*
	 *	Check for RLE sets
	 */
	len = 2+1+3;
	j = 0;
	for(i = 1; i < 28; ++i)
	{
		if(row[i] != row[j])
		{
			len += 1+5+3;
			j = i;
		}
	}
	if(len < bestlen)
	{
		besttype = 0;
		bestlen = len;
	}

	/*
	 *	last but not least, check for all-zeros-except
	 */
	FillObjectTiles(lev,y,0);	// replace objects with tile 0
	len = 3;
	for(i = 0; i < 28; ++i)
	{
		if(row[i])
			len += 3+5+1;
	}
	if(len <= bestlen)
	{
		besttype = 2;
		bestlen = len;
	}

	// if type is 0,1,3 -- need to re-replace objects with adjacent tile
	if((besttype == 0) || (besttype == 1) || (besttype == 3))
		FillObjectTiles(lev,y,1);

	return besttype;
}

void Levels::FillObjectTiles(int lev,int y,int adj)
{
	int i;
	for(i = 0; i < lvl[lev].enemies; ++i)
	{
		if(lvl[lev].enY[i] != y)		continue;
		if(adj)
		{
			if(lvl[lev].enX[i] == 0)	lvl[lev].mp[y][0] = lvl[lev].mp[y][1];
			else						lvl[lev].mp[y][lvl[lev].enX[i]] = lvl[lev].mp[y][lvl[lev].enX[i] - 1];
		}
		else							lvl[lev].mp[y][lvl[lev].enX[i]] = 0;
	}

	if(lvl[lev].plY != y)			return;
	
	if(adj)
	{
		if(lvl[lev].plX == 0)		lvl[lev].mp[y][0] = lvl[lev].mp[y][1];
		else						lvl[lev].mp[y][lvl[lev].plX] = lvl[lev].mp[y][lvl[lev].plX - 1];
	}
	else							lvl[lev].mp[y][lvl[lev].plX] = 0;
}