
#include <wx/wx.h>
#include <wx/dcbuffer.h>
#include "app.h"
#include "lodemap.h"


BEGIN_EVENT_TABLE(LodeMap, wxWindow)
	EVT_PAINT(LodeMap::OnPaint)
	EVT_LEFT_DOWN(LodeMap::OnLDown)
	EVT_LEFT_UP(LodeMap::OnLUp)
	EVT_RIGHT_DOWN(LodeMap::OnRDown)
	EVT_MOTION(LodeMap::OnMouseMove)
	EVT_MOUSE_CAPTURE_LOST(LodeMap::OnCaptureLost)
END_EVENT_TABLE()

LodeMap::LodeMap(wxWindow* parent, int id)
	: wxWindow(parent,id,wxDefaultPosition,wxDefaultSize,wxBORDER_SUNKEN)
{
	SetClientSize(448,224);

	pBmp = new wxBitmap(448,224,-1);
	pDC = new wxMemoryDC(*pBmp);
	nLev = 0;
	nTile = 0;
	bDragging = 0;
}

LodeMap::~LodeMap()
{
	delete pDC;
	delete pBmp;
}

void LodeMap::SetLevel(int lev)
{
	nLev = lev;

	int x, y;
	Levels& lv = wxGetApp().GetLevels();
	Graphic& gfx = wxGetApp().Gfx();

	for(y = 0; y < 14; ++y)
	{
		for(x = 0; x < 28; ++x)
		{
			gfx.DrawTile(lv.Get(lev,x,y),pDC,x,y,0);
		}
	}

	Refresh(0,0);
}

void LodeMap::OnPaint(wxPaintEvent& event)
{
	wxPaintDC dc(this);

	if(wxGetApp().IsLoaded())
	{
		wxRegionIterator upd(GetUpdateRegion());
		int x, y, w, h;

		while (upd)
		{
			x = upd.GetX();
			y = upd.GetY();
			w = upd.GetW();
			h = upd.GetH();

			dc.Blit(x,y,w,h,pDC,x,y);
			upd++;
		}
	}
	else
	{
		wxBrush br(wxSystemSettings::GetColour(wxSYS_COLOUR_APPWORKSPACE));

		dc.SetBackground(br);
		dc.Clear();
		dc.SetBackground(wxNullBrush);
	}
}

void LodeMap::OnRDown(wxMouseEvent& event)
{
	Lift();
	if(!wxGetApp().IsLoaded())		return;

	int x = event.GetX() / 16;
	int y = event.GetY() / 16;

	if(y < 0)		return;
	if(y >= 14)		return;
	if(x < 0)		return;
	if(x >= 28)		return;

	wxGetApp().SetTile(wxGetApp().GetLevels().Get(nLev,x,y));
}

void LodeMap::OnLDown(wxMouseEvent& event)
{
	Lift();
	if(!wxGetApp().IsLoaded())		return;

	bDragging = 1;
	nDragX = event.GetX() / 16;
	nDragY = event.GetY() / 16;

	ChangeTile();

	CaptureMouse();
}

void LodeMap::OnLUp(wxMouseEvent& event)
{
	Lift();
}

void LodeMap::OnMouseMove(wxMouseEvent& event)
{
	if(!bDragging)		return;

	int x = event.GetX() / 16;
	int y = event.GetY() / 16;

	if((x == nDragX) && (y == nDragY))		return;

	nDragX = x;
	nDragY = y;
	ChangeTile();
}

void LodeMap::OnCaptureLost(wxMouseCaptureLostEvent& event)
{
	bDragging = 0;
}

void LodeMap::Lift()
{
	if(!bDragging)		return;

	ReleaseMouse();
	bDragging = 0;
}

void LodeMap::ChangeTile()
{
	if(nDragY < 0)			return;
	if(nDragY >= 14)		return;
	if(nDragX < 0)			return;
	if(nDragX >= 28)		return;

	wxGetApp().Dirty();

	int a_x, a_y;
	a_x = -1;
	a_y = -1;

	wxGetApp().GetLevels().Set(nLev,nDragX,nDragY,nTile,a_x,a_y);

	UpdateTile(nDragX,nDragY);
	if(a_y >= 0)
		UpdateTile(a_x,a_y);
}

void LodeMap::UpdateTile(int x,int y)
{	
	wxGetApp().Gfx().DrawTile(wxGetApp().GetLevels().Get(nLev,x,y),pDC,x,y,0);

	RefreshRect(wxRect(x*16,y*16,16,16),0);
}